// module wxTest.cpp
// written by EH 01-2013, derived from wxWidgets sample "Minimal"
// 28.01.2021 corrected for newer C++
//            make file for Linux corrected (includes now wx-config)
//            for Windows/wxWidgets V3.1.0: see make file wxTestW.mak
//            for Linux/Ubuntu wxWidgets 3.?.? see make file wxTestUnx.mak
// 09.01.2022 Tests for wxString ---> UTF8 added
// 10.01.2022 Test for wxList<T> added

#include <wx/wx.h>
#include <wx/msgqueue.h>
//#include "wx/numformatter.h"  // for test only

class MyFrame;
class MyApp;

MyApp *pGlobalApp = NULL;
int logging = 0;

/* XPM */
static const char *wxTest_xpm[] = {
/* columns rows colors chars-per-pixel */
"33 34 18 1",
"  c #FFFFFF",
"o c #60C080",
"X c #C0DCC0",
"+ c #60E080",
"& c #80E080",
"$ c #00C000",
"* c #00E000",
". c #FFFBF0",
"= c #A0E080",
"% c #20E000",
"> c #20C040",
"# c #20E040",
", c #40A040",
"O c #40C040",
"- c #00FF00",
"; c #40E040",
"@ c #60C040",
": c #60E040",
/* pixels */
"    ....         ...             ",
"    ....                         ",
"              ..                 ",
"             ...                 ",
"             .........          .",
"              ........           ",
"                                 ",
"                 .               ",
"       . .  .   .X........X......",
"    .   ........oO+XOXXXOX@@X@XX.",
"...      ..X.XX@##$%#&###*&##%##X",
"...     ..XXXOO=*-#---*-*------*#",
"...  . ..X+;##;%-*--*-%*--**--*--",
"..   ...XX;*---------**--*-------",
"     ..X;#*----------------------",
"     .X:#----------*-%*-*#*-----*",
"     .X#---------*%;#;##;;;=##*%%",
"   ..X+*-----*#--%;=XXX+XXXX;>;==",
"    ..;#----*%%##:..X.X.......X:O",
"    .X>------%XXXX.     .     .XX",
"   ..X#---**%;XX..  ..    .   .. ",
"   ..X#--**#%&X  .      .  .. .  ",
"   .X:%*---**;. .    . ....   .  ",
"   .X>---*-%%= .     ........    ",
"  ..X;-------=.     ...          ",
"  ..XO-------;X  .....          .",
"  ..XO%------%X...XX..           ",
"  ...&%------%X,X@OoX..          ",
"   ..X#----*-*#=>;=O:X.   ...    ",
"    ..+#------**##&;;XX   .......",
"     .=#---------*-*#X.    ...   ",
"   . .X#------------#X.    ...   ",
"     .X;------------*;.   ....   ",
"..    .X*------------#.   ...    "
};


// -----------------------------------------------------------
class MyApp : public wxApp
{
public:
    virtual bool OnInit();
};

// -------------------------------------------------------------
class MyTimer : public wxTimer
{
    MyFrame *owner;
  public:
    MyTimer(MyFrame *pOwner, int id);
    void Notify();
    MyFrame *getOwner()
    { return this->owner; };
};

// -----------------------------------------------------------
// This is the windows where we draw on
class MyWindow : public wxScrolledWindow
{
  public:
    wxClientDC  *pDc;
    MyWindow(MyFrame *parent, wxPoint origin, wxSize size);
};

// ------------------------------------------------------------
// Define a new frame type: this is going to be our main frame
class MyFrame : public wxFrame
{
  public:
    MyFrame(const wxString& title);
    MyWindow  *pw;  
    MyTimer   *pTimer;
    
    void OnQuit(wxCommandEvent& event);
    void OnAbout(wxCommandEvent& event);
    void OnActivate(wxActivateEvent& event);
    MyWindow *getWindow()
    { return this->pw;  };    

  private:
    // any class wishing to process wxWidgets events must use this macro
    DECLARE_EVENT_TABLE()
};

enum
{
    wxTest_Quit = wxID_EXIT,
    wxTest_About = wxID_ABOUT
};

// ------------------------------------------------------------------------
BEGIN_EVENT_TABLE(MyFrame, wxFrame)
  EVT_MENU(wxTest_Quit,  MyFrame::OnQuit)
  EVT_MENU(wxTest_About, MyFrame::OnAbout)
  EVT_ACTIVATE(MyFrame::OnActivate)
END_EVENT_TABLE()

IMPLEMENT_APP(MyApp)

// -------------------------------------------------------------------------
// Tests wxList
// wxList keeps references to elements (no copy semantics!)
// Note: The wxWidgets docu recommends STL lists!
// This macro declares and partly implements MyList class
WX_DECLARE_LIST(wxString, MyList);  // list contains pointers to wxString objects
// if this is lacking: "undefined reference to vtable for MyLIstNode"
#include <wx/listimpl.cpp> // must be "cpp"!
WX_DEFINE_LIST(MyList);

void testList()
{
  MyList list1;
  wxString s1, s2, s3;
  wxString *pElement;
  //
  ::wxPrintf("testList() start\n");
  s1 = "Element1";
  s2 = "Element2";
  s3 = "Element3";
  //
  // insert elements
  list1.Append(&s1);
  list1.Append(&s2);
  list1.Append(&s3);
  ::wxPrintf("testList(): no.of list elements=%lu\n",
              (unsigned long)(list1.GetCount()));
  //
  // delete the contents (invoke destructor): this sets only a flag? Yes!
  // Makes tis sense only for allocated list objects?i Yes!
  //list1.DeleteContents(true);
  //
  // Iterate over the list elements
  // This is "NOT STL" syntax
  MyList::compatibility_iterator it = list1.GetFirst();
  while (it)
  {
    pElement = it->GetData(); 
    ::wxPrintf("testList(): list element=%s\n", *pElement);
    it = it->GetNext();
  }
  ::wxPrintf("testList() delete an object\n");
  list1.DeleteObject(&s2);
  ::wxPrintf("testList(): no.of list elements=%lu\n",
              (unsigned long)(list1.GetCount()));
  // Note: no crash even for a non-allocated object!
  // after deletion: show the contents again
  it = list1.GetFirst();
  while (it)
  {
    pElement = it->GetData(); 
    ::wxPrintf("testList(): list element=%s\n", *pElement);
    it = it->GetNext();
  }
  ::wxPrintf("testList() end\n");
}   // end testList()
  
// -------------------------------------------------------------------------
// Test conversion internal wxString to UTF8 (1-byte code)
// Note: Windows: when Umlauts appear in wxString the encoding must be "Codepage"
//       Linux: the encoding must be "UTF-8" (in scite etc)
void testUTF8()
{
  // solution for this problem: see OnInit(): wxLocale must be specified!!!
  // ERROR under Linux when Umlauts included in assignment!!! error
  wxString wxS = "Hallo ";  // 10 chars when umlauts included
  // Note: printf() statt wxPrintf() also OK in this sample
  // had to use "%ld" with cast "unsigned long" - G++ und Linux requires this
  ::wxPrintf("testUTF8(): sizeof wxS=%ld\n", (unsigned long)sizeof(wxS));              // Ausgabe: 32
  ::wxPrintf("testUTF8(): wxS.Len()=%ld\n", (unsigned long)wxS.Len());                 // Ausgabe: 10                // Ausgabe: 10
  ::wxPrintf("testUTF8(): sizeof(wxChar)=%ld\n", (unsigned long)sizeof(wxChar));       // Ausgabe : 2
  ::wxPrintf("testUTF8(): sizeof(wxUniChar)=%ld\n", (unsigned long)sizeof(wxUniChar)); // Ausgabe : 4
  //
  const char *p;
  p = wxS.ToUTF8();
  ::wxPrintf("Hallo C-String: %s\n", p); 
  // ToUTF8(): 1-Byte fuer ASCII Zeichen, 2 Bytes fuer Umlaute etc.
  // hier strlen(p) ist 14, wegen 2-Bytes Kodierung fuer Umlaute!!!!
  ::wxPrintf("Laenge(strlen) C-String(UTF-8): %ld\n", 
             (unsigned long)::strlen(p)); // Ausgabe: 14
  // c_str(): erzwingt 1-Byte characters
  p = wxS.c_str();
  ::wxPrintf("Laenge(strlen) gleicher C-String(c_str()): %ld\n", 
             (unsigned long)::strlen(p)); // Ausgabe: 10
  //
  ::wxPrintf("sizeof wxString(Hallo + 4 umlaute): %ld\n", 
              sizeof(wxS));   // Ausgabe: 32: 10*2 Bytes + 12 Bytes Overhead
  //
  wxChar wxC;
  wxUniChar wxUniC;
  for (int n=0; n < wxS.Len(); n++)
  {
    wxC    = wxS[n];
    wxUniC = wxS[n];
    ::wxPrintf("testUTF8(): wxC=%c wxUniC=%c\n", wxC, wxUniC);
  }
  
  ::wxPrintf("testUTF8(): end\n");
}   // end testUTF8()

// -------------------------------------------------------------------------
// Test wxWidgets queues: recommended for communication between threads
// As we don't used threads here this is not very realistic
// Result shows that this is a FIFO queue
void testQ()
{
  wxMessageQueue<wxString> q1;
  wxString s1;
  wxMessageQueueError err;
  //
  ::wxPrintf("testQ(): start\n");
  q1.Post(wxString("123456"));
  q1.Post(wxString("abcde"));
  q1.Post(wxString("VWXYZ"));
  //
  do
  {
    //err = q1.Receive(s1);  // this blocks if no more entry in the queue
    err = q1.ReceiveTimeout(5, s1); // returns wxMSGQUEUE_TIMEOUT if no entry
    if (err == wxMSGQUEUE_NO_ERROR)
      ::wxPrintf("from queue: " + s1 + "\n");
  } while (err == wxMSGQUEUE_NO_ERROR);
  //
  ::wxPrintf("testQ(): end\n");
}   // end testQ()

// ------------------------------------------------------------------------
// 'Main program' equivalent: the program execution "starts" here
bool MyApp::OnInit()
{
  if (! wxApp::OnInit())
    return false;
  // solution for Umlaut problems under Linux
  // Can also be used under Windows 32-bit
  // see http://trilix.de/ver-02/index.php/programmieren/wxwidgets/umlautproblem-bei-den-wx-controls-loesen-linux
  wxLocale* locale = new wxLocale();
  locale->Init(wxLANGUAGE_GERMAN, wxLOCALE_LOAD_DEFAULT);
  //
  ::pGlobalApp = this;
  MyFrame *frame = new MyFrame(_T("wxTest App"));
  frame->Show(true);
  //
  ::testUTF8();
  //
  ::testList();
  //
  ::testQ();
  //
  return true;
}

// ------------------------------------------------------------
// constructor
MyWindow::MyWindow(MyFrame *parent, 
                   wxPoint origin,
                   wxSize size) : wxScrolledWindow(parent, -1, origin, size)
{
  this->pDc = NULL; 
}

// ------------------------------------------------------------
// constructor main frame
MyFrame::MyFrame(const wxString& title)
       : wxFrame(NULL, wxID_ANY, title)
{
  
  // set the frame icon
  SetIcon(wxICON(wxTest));
  this->pw = NULL;
  
  // create a menu bar
  wxMenu *fileMenu = new wxMenu;

  // the "About" item should be in the help menu
  wxMenu *helpMenu = new wxMenu;
  helpMenu->Append(wxTest_About,
                    _T("&About...\tF1"), _T("Show about dialog"));
  fileMenu->Append(wxTest_Quit, 
                   _T("E&xit\tAlt-X"), _T("Quit this program"));
  // now append the freshly created menu to the menu bar...
  wxMenuBar *menuBar = new wxMenuBar();
  menuBar->Append(fileMenu, _T("&File"));
  menuBar->Append(helpMenu, _T("&Help"));

  // ... and attach this menu bar to the frame
  SetMenuBar(menuBar);

  // create a status bar just for fun (by default with 1 pane only)
  CreateStatusBar(2);
  SetStatusText(_T("Welcome to wxWidgets!"));
  //
  wxPoint point(0, 0);
  wxSize size   = this->GetClientSize();
  this->pw = new MyWindow(this, point, size);
  //
  this->pTimer = new MyTimer(this, 
                             ::wxNewId());
   //
   // Calls Notify() every 250 ms
  this->pTimer->Start(250, 
                      wxTIMER_CONTINUOUS);
}   // end frame constructor

// ----------------------------------------------------------------
void MyFrame::OnActivate(wxActivateEvent& event)
{
  if (this->pw != NULL)
  {
    try
    {
      if (event.GetActive())
      {
        // Fenster wird aktiviert
        //wxClientDC dc(this->pw);   // crash!!!!!!!!
      }
      else
      {
        // wird deaktiviert
      }
    }
    catch (...)
    {
    }
  }
  event.Skip();
}   // end MyFrame::OnActivate()

// --------------------------------------------------------------
// event handlers
void MyFrame::OnQuit(wxCommandEvent& WXUNUSED(event))
{
    // true is to force the frame to close
    Close(true);
}

// ---------------------------------------------------------------
void MyFrame::OnAbout(wxCommandEvent& WXUNUSED(event))
{
    wxMessageBox(wxString::Format(
                    _T("Welcome to %s!\n")
                    _T("\n")
                    _T("This is the wxTest app\n")
                    _T("running under %s."),
                    wxVERSION_STRING,
                    wxGetOsDescription().c_str()
                 ),
                 _T("About wxTest"),
                 wxOK | wxICON_INFORMATION,
                 this);
}   // end MyFrame::OnAbout()

// ----------------------------------------------------------------
// Constructor Timer
MyTimer::MyTimer(MyFrame *pOwner, int id) :
                   wxTimer(pOwner, id)
{
  this->owner = pOwner;
}

// ---------------------------------------------------------------
void drawFilledRectangle(wxClientDC *pdc,
                         int xp, int yp,
                         int xl, int yl)
{
  static int count = 0;
  wxMemoryDC memDC;
  wxBitmap bitmap(xl, yl);
  memDC.SelectObject(bitmap);
  if ((count % 2) == 0)
    memDC.SetBackground(*wxCYAN_BRUSH); //*wxWHITE_BRUSH);
  else
    memDC.SetBackground(*wxGREY_BRUSH);
  memDC.Clear();   // must be!!!
  pdc->Blit(xp, yp, xl, yl, &memDC, 0, 0, wxCOPY); 
  // wxSET can be used to erase the background (white background)
  count++;
  return;
}   // end drawFilledRectangle()

// ---------------------------------------------------------------
// Callback Routine for the timer
void MyTimer::Notify()
{
  wxPen *pp       = NULL;
  wxBrush *pb     = NULL;
  MyFrame *pFrame = NULL;
  //wxScreenDC sdc;   //    ist OK - but for the complete screen
  wxClientDC *pWdc = NULL;   
  MyWindow  *pWin  = NULL;
  int        width, height;
  wxSize size;
  int verlauf           = 0; 
  static int yp         = 10;
  // count counts the overflows in the drawing routines
  static unsigned count = 0;
  //
  if (::logging) 
    printf("MyTimer::Notify() count=%u start\n", count);
  if (::pGlobalApp == NULL)
    goto zurueck;
  verlauf = 1;
  pFrame = this->getOwner();
  if (pFrame == NULL)
    goto zurueck;
  //
  verlauf = 2;
  pWin = pFrame->getWindow();
  if (pWin == NULL)
    goto zurueck;
  //
  verlauf = 3;
  pWdc = new wxClientDC(pWin);
  if (pWdc == NULL)
    goto zurueck;
  //
  verlauf = 4;
  size   = pWin->GetClientSize();
  width  = size.GetWidth();
  height = size.GetHeight();
  if (::logging) 
    printf("MyTimer::Notify() width=%d height=%d\n", width, height);
  if ((count % 3) == 0)
    pp = new wxPen(wxColour(255, 0, 0), 2, wxPENSTYLE_SOLID);  // RGB red
  if ((count % 3) == 1)
    pp = new wxPen(wxColour(0, 255, 0), 2, wxPENSTYLE_SOLID);  // RGB green
  if ((count % 3) == 2)
  {   
    verlauf = 5;
    ::drawFilledRectangle(pWdc, 40, yp, width - 80, 50);
  }
  else
  {
    verlauf = 6;
    pb = new wxBrush("BLACK", wxBRUSHSTYLE_SOLID);
    pWdc->SetBrush(*pb);
    pWdc->SetPen(*pp);
    pWdc->DrawLine(0, yp, width, yp);  // horizontale Linie
  }
  if (::logging) 
    printf("MyTimer::Notify() WindowDC() ok\n");
  //
  verlauf = 6;
  yp = yp + 10;
  if (yp > height)
  {
    count++;
    yp = 10;
  }
  //
  zurueck:
  if (pp != NULL)   delete pp;
  if (pb != NULL)   delete pb;
  if (pWdc != NULL) delete pWdc;
  if (::logging) 
    printf("MyTimer::Notify() verlauf=%d end\n", verlauf);
}   // end MyTimer::Notify()
